@@ -641,28 +641,38 @@ mod treeindex_test {
             }
         }
 
-        let data_size = 65536; // 1_048_576;
+        let data_size = 1_048_576;
         let tree: TreeIndex<usize, R> = TreeIndex::new();
         for k in 0..data_size {
             assert!(tree.insert(k, R::new()).is_ok());
         }
-        for k in 0..data_size {
+        for k in (0..data_size).rev() {
             assert!(tree.remove(&k));
         }
+
+        let mut cnt = 0;
         while INST_CNT.load(Relaxed) > 0 {
             let barrier = ebr::Barrier::new();
             drop(barrier);
+            cnt += 1;
         }
+        println!("{cnt}");
+        assert!(cnt >= INST_CNT.load(Relaxed));
 
         let tree: TreeIndex<usize, R> = TreeIndex::new();
-        for k in 0..data_size {
+        for k in 0..(data_size / 16) {
             assert!(tree.insert(k, R::new()).is_ok());
         }
         tree.clear();
+
+        let mut cnt = 0;
         while INST_CNT.load(Relaxed) > 0 {
             let barrier = ebr::Barrier::new();
             drop(barrier);
+            cnt += 1;
         }
+        println!("{cnt}");
+        assert!(cnt >= INST_CNT.load(Relaxed));
     }
 
     #[test]
@@ -766,7 +776,17 @@ mod treeindex_test {
                         assert_eq!(entry, (&(first_key + 3), &(first_key + 3)));
                     }
 
+                    let key_at_halfway = first_key + range / 2;
                     for key in (first_key + 1)..(first_key + range) {
+                        if key == key_at_halfway {
+                            let ebr_barrier = ebr::Barrier::new();
+                            let mut range_scanner =
+                                tree_copied.range((first_key + 1).., &ebr_barrier);
+                            let entry = range_scanner.next().unwrap();
+                            assert_eq!(entry, (&key_at_halfway, &key_at_halfway));
+                            let entry = range_scanner.next().unwrap();
+                            assert_eq!(entry, (&(key_at_halfway + 1), &(key_at_halfway + 1)));
+                        }
                         assert!(tree_copied.remove(&key));
                         assert!(!tree_copied.remove(&key));
                         assert!(tree_copied.read(&(first_key + 1), |_, _| ()).is_none());
