@@ -28,6 +28,9 @@ pub enum InsertResult<K, V> {
     ///
     /// It is a terminal state.
     Retired(K, V),
+
+    /// The operation can be retried.
+    Retry(K, V),
 }
 
 /// The result of removal.
@@ -36,11 +39,17 @@ pub enum RemoveResult {
     /// Remove succeeded.
     Success,
 
-    /// Remove failed.
-    Fail,
+    /// Remove succeeded and cleanup required.
+    Cleanup,
 
     /// Remove succeeded and the [`Leaf`] has retired without usable entries left.
     Retired,
+
+    /// Remove failed.
+    Fail,
+
+    /// The [`Leaf`] is frozen.
+    Frozen,
 }
 
 /// The number of entries and number of state bits per entry.
@@ -181,11 +190,6 @@ where
         }
     }
 
-    /// Returns `true` if the [`Leaf`] is frozen.
-    pub fn frozen(&self) -> bool {
-        Dimension::frozen(self.metadata.load(Relaxed))
-    }
-
     /// Thaws the [`Leaf`].
     pub fn thaw(&self) -> bool {
         self.metadata
@@ -272,6 +276,9 @@ where
         Q: Ord + ?Sized,
     {
         let mut metadata = self.metadata.load(Acquire);
+        if Dimension::frozen(metadata) {
+            return RemoveResult::Frozen;
+        }
         let mut max_min_rank = 0;
         let mut min_max_rank = DIMENSION.removed_state();
         for i in 0..DIMENSION.num_entries {
@@ -330,6 +337,9 @@ where
                                     if DIMENSION.state(actual, i) == DIMENSION.removed_state() {
                                         return RemoveResult::Fail;
                                     }
+                                    if Dimension::frozen(actual) {
+                                        return RemoveResult::Frozen;
+                                    }
                                     metadata = actual;
                                 }
                             }
@@ -469,9 +479,22 @@ where
         (usize::MAX, ptr::null())
     }
 
-    /// Freezes the [`Leaf`] and distribute entries to two new leaves.
+    /// Freezes the [`Leaf`] temporarily.
     ///
     /// A frozen [`Leaf`] cannot store more entries, and on-going insertion is cancelled.
+    pub fn freeze(&self) -> bool {
+        self.metadata
+            .fetch_update(AcqRel, Acquire, |p| {
+                if Dimension::frozen(p) {
+                    None
+                } else {
+                    Some(Dimension::freeze(p))
+                }
+            })
+            .is_ok()
+    }
+
+    /// Freezes the [`Leaf`] and distribute entries to two new leaves.
     pub fn freeze_and_distribute(
         &self,
         low_key_leaf: &mut Option<Arc<Leaf<K, V>>>,
@@ -902,13 +925,14 @@ mod test {
         assert_eq!(leaf1.as_ref().and_then(|l| l.search(&17)), Some(&11));
         assert!(leaf2.is_none());
         assert!(matches!(leaf.insert(1, 7), InsertResult::Frozen(..)));
-        assert_eq!(leaf.remove_if(&17, &mut |_| true), RemoveResult::Success);
+        assert_eq!(leaf.remove_if(&17, &mut |_| true), RemoveResult::Frozen);
         assert!(matches!(leaf.insert(3, 5), InsertResult::Frozen(..)));
 
         assert!(leaf.thaw());
         assert!(matches!(leaf.insert(1, 7), InsertResult::Success));
 
         assert_eq!(leaf.remove_if(&1, &mut |_| true), RemoveResult::Success);
+        assert_eq!(leaf.remove_if(&17, &mut |_| true), RemoveResult::Success);
         assert_eq!(leaf.remove_if(&11, &mut |_| true), RemoveResult::Retired);
 
         assert!(matches!(leaf.insert(5, 3), InsertResult::Retired(..)));
@@ -982,7 +1006,8 @@ mod test {
                         }
                         InsertResult::Duplicate(_, _)
                         | InsertResult::Frozen(_, _)
-                        | InsertResult::Retired(_, _) => {
+                        | InsertResult::Retired(_, _)
+                        | InsertResult::Retry(_, _) => {
                             unreachable!();
                         }
                         InsertResult::Full(k, v) => {
@@ -1016,6 +1041,7 @@ mod test {
                     match leaf_clone.remove_if(&t, &mut |_| true) {
                         RemoveResult::Success => assert!(inserted),
                         RemoveResult::Fail => assert!(!inserted),
+                        RemoveResult::Frozen | RemoveResult::Cleanup => unreachable!(),
                         RemoveResult::Retired => {
                             assert!(inserted);
                             assert_eq!(retire_clone.swap(1, Relaxed), 0);
