@@ -8,6 +8,7 @@ use crate::LinkedList;
 use std::borrow::Borrow;
 use std::cmp::Ordering::{Equal, Greater, Less};
 use std::ptr::addr_of;
+use std::sync::atomic::AtomicPtr;
 use std::sync::atomic::Ordering::{self, AcqRel, Acquire, Relaxed, Release};
 
 /// [`Tag::First`] indicates the corresponding node has retired.
@@ -196,24 +197,25 @@ where
                         // Data race resolution - see `LeafNode::search`.
                         match child_ref.insert(key, value) {
                             InsertResult::Success => return Ok(InsertResult::Success),
-                            InsertResult::Duplicate(key, value) => {
-                                return Ok(InsertResult::Duplicate(key, value));
+                            InsertResult::Duplicate(k, v) => {
+                                return Ok(InsertResult::Duplicate(k, v));
                             }
-                            InsertResult::Full(key, value) | InsertResult::Retired(key, value) => {
+                            InsertResult::Full(k, v) | InsertResult::Retired(k, v) => {
                                 return self.split_leaf::<ASYNC>(
-                                    key,
-                                    value,
+                                    k,
+                                    v,
                                     Some(child_key),
                                     child_ptr,
                                     child,
                                     barrier,
                                 );
                             }
-                            InsertResult::Frozen(key, value) => {
+                            InsertResult::Frozen(k, v) => {
                                 // The `Leaf` is being split: retry.
-                                self.wait::<ASYNC>(true, barrier);
-                                return Err((key, value));
+                                self.wait::<ASYNC>(barrier);
+                                return Err((k, v));
                             }
+                            InsertResult::Retry(..) => unreachable!(),
                         };
                     }
                 }
@@ -245,23 +247,24 @@ where
                 }
                 match unbounded.insert(key, value) {
                     InsertResult::Success => return Ok(InsertResult::Success),
-                    InsertResult::Duplicate(key, value) => {
-                        return Ok(InsertResult::Duplicate(key, value));
+                    InsertResult::Duplicate(k, v) => {
+                        return Ok(InsertResult::Duplicate(k, v));
                     }
-                    InsertResult::Full(key, value) | InsertResult::Retired(key, value) => {
+                    InsertResult::Full(k, v) | InsertResult::Retired(k, v) => {
                         return self.split_leaf::<ASYNC>(
-                            key,
-                            value,
+                            k,
+                            v,
                             None,
                             unbounded_ptr,
                             &self.unbounded_child,
                             barrier,
                         );
                     }
-                    InsertResult::Frozen(key, value) => {
-                        self.wait::<ASYNC>(true, barrier);
-                        return Err((key, value));
+                    InsertResult::Frozen(k, v) => {
+                        self.wait::<ASYNC>(barrier);
+                        return Err((k, v));
                     }
+                    InsertResult::Retry(..) => unreachable!(),
                 };
             }
             return Ok(InsertResult::Retired(key, value));
@@ -291,14 +294,13 @@ where
                     if self.children.validate(metadata) {
                         // Data race resolution - see `LeafNode::search`.
                         let result = child.remove_if(key, condition);
-                        if child.frozen() {
+                        if result == RemoveResult::Frozen {
                             // When a `Leaf` is frozen, its entries may be being copied to new
                             // `Leaves`.
-                            self.wait::<ASYNC>(true, barrier);
-                            return Err(result != RemoveResult::Fail);
-                        }
-                        if result == RemoveResult::Retired {
-                            return Ok(self.coalesce(key, barrier));
+                            self.wait::<ASYNC>(barrier);
+                            return Err(false);
+                        } else if result == RemoveResult::Retired {
+                            return Ok(self.coalesce(barrier));
                         }
                         return Ok(result);
                     }
@@ -314,12 +316,11 @@ where
                     continue;
                 }
                 let result = unbounded.remove_if(key, condition);
-                if unbounded.frozen() {
-                    self.wait::<ASYNC>(true, barrier);
-                    return Err(result != RemoveResult::Fail);
-                }
-                if result == RemoveResult::Retired {
-                    return Ok(self.coalesce(key, barrier));
+                if result == RemoveResult::Frozen {
+                    self.wait::<ASYNC>(barrier);
+                    return Err(false);
+                } else if result == RemoveResult::Retired {
+                    return Ok(self.coalesce(barrier));
                 }
                 return Ok(result);
             }
@@ -350,6 +351,8 @@ where
                     origin_leaf: full_leaf.clone(Relaxed, barrier),
                     low_key_leaf: AtomicArc::null(),
                     high_key_leaf: AtomicArc::null(),
+                    low_key_leaf_node: AtomicPtr::default(),
+                    high_key_leaf_node: AtomicPtr::default(),
                 })),
                 Tag::None,
             ),
@@ -359,7 +362,7 @@ where
         ) {
             ptr.as_ref().unwrap()
         } else {
-            self.wait::<ASYNC>(false, barrier);
+            self.wait::<ASYNC>(barrier);
             return Err((key, value));
         };
 
@@ -412,7 +415,55 @@ where
         // having been `scanned`.
         let low_key_leaf_ptr = new_leaves.low_key_leaf.load(Relaxed, barrier);
         let high_key_leaf_ptr = new_leaves.high_key_leaf.load(Relaxed, barrier);
-        let unused_leaf = if high_key_leaf_ptr.is_null() {
+        let unused_leaf = if let Some(high_key_leaf) = high_key_leaf_ptr.as_ref() {
+            // From here, `Scanners` can reach the new leaves.
+            let result =
+                target.push_back(high_key_leaf_ptr.get_arc().unwrap(), true, Release, barrier);
+            debug_assert!(result.is_ok());
+            let result =
+                target.push_back(low_key_leaf_ptr.get_arc().unwrap(), true, Release, barrier);
+            debug_assert!(result.is_ok());
+
+            // Take the max key value stored in the low key leaf as the leaf key.
+            let low_key_leaf = low_key_leaf_ptr.as_ref().unwrap();
+            let max_key = low_key_leaf.max().unwrap().0;
+
+            // Need to freeze the leaf before trying to make it reachable.
+            let frozen = low_key_leaf.freeze();
+            debug_assert!(frozen);
+
+            match self.children.insert(
+                max_key.clone(),
+                new_leaves.low_key_leaf.clone(Relaxed, barrier),
+            ) {
+                InsertResult::Success => (),
+                InsertResult::Duplicate(..)
+                | InsertResult::Frozen(..)
+                | InsertResult::Retry(..) => unreachable!(),
+                InsertResult::Full(_, _) | InsertResult::Retired(_, _) => {
+                    // Need to freeze the other leaf.
+                    let frozen = high_key_leaf.freeze();
+                    debug_assert!(frozen);
+                    return Ok(InsertResult::Full(key, value));
+                }
+            };
+
+            // Mark the full leaf deleted before making the new one reachable and updatable.
+            //
+            // If the order is reversed, there emerges a possibility that entries were removed from
+            // the replaced leaf node whereas those entries still remain in `unused_leaf`; if that
+            // happens, iterators may see the removed entries momentarily.
+            let delete = target.delete_self(Release);
+            debug_assert!(delete);
+
+            // Unfreeze the leaf.
+            let unfrozen = low_key_leaf.thaw();
+            debug_assert!(unfrozen);
+
+            // Replace the full leaf with the high-key leaf.
+            let high_key_leaf = new_leaves.high_key_leaf.swap((None, Tag::None), Relaxed).0;
+            full_leaf.swap((high_key_leaf, Tag::None), Release).0
+        } else {
             // From here, `Scanners` can reach the new leaf.
             //
             // The full leaf is marked so that readers know that the next leaves may contain
@@ -422,10 +473,6 @@ where
             debug_assert!(result.is_ok());
 
             // Mark the full leaf deleted before making the new one reachable and updatable.
-            //
-            // There is a possibility that entries were removed from the replaced leaf node whereas
-            // those entries still remain in `unused_leaf`; if that happens, iterators may see
-            // the removed entries momentarily.
             let deleted = target.delete_self(Release);
             debug_assert!(deleted);
 
@@ -435,35 +482,6 @@ where
                     Release,
                 )
                 .0
-        } else {
-            // From here, `Scanners` can reach the new leaves.
-            let result =
-                target.push_back(high_key_leaf_ptr.get_arc().unwrap(), true, Release, barrier);
-            debug_assert!(result.is_ok());
-            let result =
-                target.push_back(low_key_leaf_ptr.get_arc().unwrap(), true, Release, barrier);
-            debug_assert!(result.is_ok());
-
-            // Takes the max key value stored in the low key leaf as the leaf key.
-            let max_key = low_key_leaf_ptr.as_ref().unwrap().max().unwrap().0;
-            match self.children.insert(
-                max_key.clone(),
-                new_leaves.low_key_leaf.clone(Relaxed, barrier),
-            ) {
-                InsertResult::Success => (),
-                InsertResult::Duplicate(..) | InsertResult::Frozen(..) => unreachable!(),
-                InsertResult::Full(_, _) | InsertResult::Retired(_, _) => {
-                    return Ok(InsertResult::Full(key, value))
-                }
-            };
-
-            // Mark the full leaf deleted before making the new one reachable and updatable.
-            let delete = target.delete_self(Release);
-            debug_assert!(delete);
-
-            // Replace the full leaf with the high-key leaf.
-            let high_key_leaf = new_leaves.high_key_leaf.swap((None, Tag::None), Relaxed).0;
-            full_leaf.swap((high_key_leaf, Tag::None), Release).0
         };
 
         // Unlocks the leaf node.
@@ -477,11 +495,8 @@ where
             barrier.reclaim(unused_leaf);
         }
 
-        // Traverse several leaf nodes in order to cleanup deprecated links.
-        self.max_le_appr(key.borrow(), barrier);
-
         // Since a new leaf has been inserted, the caller can retry.
-        Err((key, value))
+        Ok(InsertResult::Retry(key, value))
     }
 
     /// Splits itself into the given leaf nodes, and returns the middle key value.
@@ -492,40 +507,47 @@ where
         barrier: &Barrier,
     ) -> Option<K> {
         let mut middle_key = None;
+        let new_leaves = self.latch.load(Relaxed, barrier).as_ref().unwrap();
 
-        debug_assert!(!self.latch.load(Relaxed, barrier).is_null());
-        let new_leaves_ref = self.latch.load(Relaxed, barrier).as_ref().unwrap();
+        low_key_leaf_node.latch.swap((None, LOCKED), Relaxed);
+        high_key_leaf_node.latch.swap((None, LOCKED), Relaxed);
+
+        // It is safe to keep the pointers to the new leaf nodes in this full leaf node since the
+        // whole split operation is protected under a single `ebr::Barrier`, and the pointers are
+        // only dereferenced during the operation.
+        new_leaves
+            .low_key_leaf_node
+            .swap(addr_of!(*low_key_leaf_node) as *mut _, Relaxed);
+        new_leaves
+            .high_key_leaf_node
+            .swap(addr_of!(*high_key_leaf_node) as *mut _, Relaxed);
 
         // Builds a list of valid leaves
         #[allow(clippy::type_complexity)]
         let mut entry_array: [Option<(Option<&K>, AtomicArc<Leaf<K, V>>)>;
             DIMENSION.num_entries + 2] = Default::default();
         let mut num_entries = 0;
-        let low_key_leaf_ref = new_leaves_ref
+        let low_key_leaf_ref = new_leaves
             .low_key_leaf
             .load(Relaxed, barrier)
             .as_ref()
             .unwrap();
         let middle_key_ref = low_key_leaf_ref.max().unwrap().0;
         for entry in Scanner::new(&self.children) {
-            if new_leaves_ref
+            if new_leaves
                 .origin_leaf_key
                 .as_ref()
                 .map_or_else(|| false, |key| entry.0.borrow() == key)
             {
                 entry_array[num_entries].replace((
                     Some(middle_key_ref),
-                    new_leaves_ref.low_key_leaf.clone(Relaxed, barrier),
+                    new_leaves.low_key_leaf.clone(Relaxed, barrier),
                 ));
                 num_entries += 1;
-                if !new_leaves_ref
-                    .high_key_leaf
-                    .load(Relaxed, barrier)
-                    .is_null()
-                {
+                if !new_leaves.high_key_leaf.load(Relaxed, barrier).is_null() {
                     entry_array[num_entries].replace((
                         Some(entry.0),
-                        new_leaves_ref.high_key_leaf.clone(Relaxed, barrier),
+                        new_leaves.high_key_leaf.clone(Relaxed, barrier),
                     ));
                     num_entries += 1;
                 }
@@ -535,7 +557,7 @@ where
             }
         }
         #[allow(clippy::branches_sharing_code)]
-        if new_leaves_ref.origin_leaf_key.is_some() {
+        if new_leaves.origin_leaf_key.is_some() {
             // If the origin is a bounded node, assign the unbounded leaf as the high key node's
             // unbounded.
             entry_array[num_entries].replace((None, self.unbounded_child.clone(Relaxed, barrier)));
@@ -545,16 +567,12 @@ where
             // unbounded.
             entry_array[num_entries].replace((
                 Some(middle_key_ref),
-                new_leaves_ref.low_key_leaf.clone(Relaxed, barrier),
+                new_leaves.low_key_leaf.clone(Relaxed, barrier),
             ));
             num_entries += 1;
-            if !new_leaves_ref
-                .high_key_leaf
-                .load(Relaxed, barrier)
-                .is_null()
-            {
+            if !new_leaves.high_key_leaf.load(Relaxed, barrier).is_null() {
                 entry_array[num_entries]
-                    .replace((None, new_leaves_ref.high_key_leaf.clone(Relaxed, barrier)));
+                    .replace((None, new_leaves.high_key_leaf.clone(Relaxed, barrier)));
                 num_entries += 1;
             }
         }
@@ -598,51 +616,81 @@ where
 
     /// Commits an on-going structural change.
     pub fn commit(&self, barrier: &Barrier) {
+        // Unfreeze both leaves.
+        if let Some(change) = self.latch.load(Relaxed, barrier).as_ref() {
+            if let Some(origin_leaf) = change.origin_leaf.swap((None, Tag::None), Relaxed).0 {
+                // Make the origin leaf unreachable before making the new leaves updatable.
+                origin_leaf.delete_self(Relaxed);
+                barrier.reclaim(origin_leaf);
+            }
+            let low_key_leaf = change.low_key_leaf.load(Relaxed, barrier).as_ref().unwrap();
+            let high_key_leaf = change
+                .high_key_leaf
+                .load(Relaxed, barrier)
+                .as_ref()
+                .unwrap();
+            let unfrozen = low_key_leaf.thaw();
+            debug_assert!(unfrozen);
+            let unfrozen = high_key_leaf.thaw();
+            debug_assert!(unfrozen);
+
+            // It is safe to dereference those pointers since the whole split operation is under a
+            // single `ebr::Barrier`.
+            if let Some(low_key_leaf_node) =
+                unsafe { change.low_key_leaf_node.load(Relaxed).as_ref() }
+            {
+                let locked = low_key_leaf_node.latch.swap((None, Tag::None), Relaxed).1;
+                debug_assert_eq!(locked, LOCKED);
+                low_key_leaf_node.wait_queue.signal();
+            }
+
+            if let Some(high_key_leaf_node) =
+                unsafe { change.high_key_leaf_node.load(Relaxed).as_ref() }
+            {
+                let locked = high_key_leaf_node.latch.swap((None, Tag::None), Relaxed).1;
+                debug_assert_eq!(locked, LOCKED);
+                high_key_leaf_node.wait_queue.signal();
+            }
+        }
+
         // Mark the leaf node retired to prevent further locking attempts.
         let (change, _) = self.latch.swap((None, RETIRED), Release);
         self.wait_queue.signal();
         if let Some(change) = change {
-            if let Some(obsolete_leaf) = change.origin_leaf.swap((None, Tag::None), Relaxed).0 {
-                // Makes the leaf unreachable before dropping it.
-                obsolete_leaf.delete_self(Relaxed);
-                barrier.reclaim(obsolete_leaf);
-            }
+            barrier.reclaim(change);
         }
     }
 
     /// Rolls back the ongoing split operation.
     pub fn rollback(&self, barrier: &Barrier) {
         if let Some(change) = self.latch.load(Relaxed, barrier).as_ref() {
-            let low_key_leaf_ptr = change.low_key_leaf.load(Relaxed, barrier);
-            let high_key_leaf_ptr = change.high_key_leaf.load(Relaxed, barrier);
+            let low_key_leaf = change.low_key_leaf.load(Relaxed, barrier).as_ref().unwrap();
+            let high_key_leaf = change
+                .high_key_leaf
+                .load(Relaxed, barrier)
+                .as_ref()
+                .unwrap();
 
             // Roll back the linked list state.
             //
             // `high_key_leaf` must be deleted first in order for `Scanners` not to omit entries.
-            if let Some(leaf_ref) = high_key_leaf_ptr.as_ref() {
-                let deleted = leaf_ref.delete_self(Relaxed);
-                debug_assert!(deleted);
-            }
-            if let Some(leaf_ref) = low_key_leaf_ptr.as_ref() {
-                let deleted = leaf_ref.delete_self(Release);
-                debug_assert!(deleted);
-            }
+            let deleted = high_key_leaf.delete_self(Relaxed);
+            debug_assert!(deleted);
+            let deleted = low_key_leaf.delete_self(Relaxed);
+            debug_assert!(deleted);
 
-            if let Some(origin) = change.origin_leaf.swap((None, Tag::None), Relaxed).0 {
-                // Thaw the [`Leaf`].
-                let result = origin.thaw();
-                debug_assert!(result);
+            if let Some(origin_leaf) = change.origin_leaf.swap((None, Tag::None), Relaxed).0 {
+                // Unfreeze the origin.
+                let unfrozen = origin_leaf.thaw();
+                debug_assert!(unfrozen);
 
                 // Remove the mark from the full leaf node.
-                //
-                // This unmarking has to be a release-store, otherwise it can be re-ordered
-                // before previous `delete_self` calls.
-                let unmarked = origin.unmark(Release);
+                let unmarked = origin_leaf.unmark(Release);
                 debug_assert!(unmarked);
             }
         }
 
-        // Unlocks the leaf node.
+        // Unlock the leaf node.
         let (change, _) = self.latch.swap((None, Tag::None), Release);
         self.wait_queue.signal();
         if let Some(change) = change {
@@ -650,20 +698,49 @@ where
         }
     }
 
+    /// Cleans up logically deleted [`LeafNode`] instances in the linked list.
+    ///
+    /// If the target leaf node does not exist in the sub-tree, returns `false`.
+    pub fn cleanup_link<'b, Q>(&self, key: &Q, tranverse_max: bool, barrier: &'b Barrier) -> bool
+    where
+        K: 'b + Borrow<Q>,
+        Q: Ord + ?Sized,
+    {
+        let scanner = if tranverse_max {
+            if let Some(unbounded) = self.unbounded_child.load(Acquire, barrier).as_ref() {
+                Scanner::new(unbounded)
+            } else {
+                return false;
+            }
+        } else if let Some(leaf_scanner) = Scanner::max_less(&self.children, key) {
+            if let Some((_, child)) = leaf_scanner.get() {
+                if let Some(child) = child.load(Acquire, barrier).as_ref() {
+                    Scanner::new(child)
+                } else {
+                    return false;
+                }
+            } else {
+                return false;
+            }
+        } else {
+            return false;
+        };
+
+        // It *would* be the maximum leaf node among those that containing keys smaller than the
+        // target key. Hopefully, two jumps will be sufficient.
+        scanner.jump(None, barrier).map(|s| s.jump(None, barrier));
+        true
+    }
+
     /// Waits for the lock on the [`LeafNode`] to be released.
-    pub(super) fn wait<const ASYNC: bool>(&self, expect_null: bool, barrier: &Barrier) {
+    pub(super) fn wait<const ASYNC: bool>(&self, barrier: &Barrier) {
         if ASYNC {
             // Currently, asynchronous waiting is not implemented.
             return;
         }
         let _result = self.wait_queue.wait(|| {
             let ptr = self.latch.load(Relaxed, barrier);
-            let is_null = ptr.is_null();
-            if expect_null && is_null {
-                // The `LeafNode` may be locked, but it does not care.
-                return Ok(());
-            }
-            if !is_null || ptr.tag() == LOCKED {
+            if !ptr.is_null() || ptr.tag() == LOCKED {
                 // The `LeafNode` is being split or locked.
                 return Err(());
             }
@@ -672,13 +749,15 @@ where
     }
 
     /// Tries to coalesce empty or obsolete leaves after a successful removal of an entry.
-    fn coalesce<Q>(&self, removed_key: &Q, barrier: &Barrier) -> RemoveResult
+    fn coalesce<Q>(&self, barrier: &Barrier) -> RemoveResult
     where
         K: Borrow<Q>,
         Q: Ord + ?Sized,
     {
+        let mut uncleaned_leaf = false;
+        let mut prev_valid_leaf = None;
         while let Some(lock) = Locker::try_lock(self, barrier) {
-            let mut has_valid_leaf = false;
+            prev_valid_leaf.take();
             for entry in Scanner::new(&self.children) {
                 let leaf_ptr = entry.1.load(Relaxed, barrier);
                 let leaf = leaf_ptr.as_ref().unwrap();
@@ -692,14 +771,20 @@ where
                     // that readers are able to retry when they find it being `null`.
                     if let Some(leaf) = entry.1.swap((None, Tag::None), Release).0 {
                         barrier.reclaim(leaf);
+                        if let Some(prev_leaf) = prev_valid_leaf.as_ref() {
+                            // One jump is sufficient.
+                            Scanner::new(*prev_leaf).jump(None, barrier);
+                        } else {
+                            uncleaned_leaf = true;
+                        }
                     }
                 } else {
-                    has_valid_leaf = true;
+                    prev_valid_leaf.replace(leaf);
                 }
             }
 
             // The unbounded leaf becomes unreachable after all the other leaves are gone.
-            let fully_empty = if has_valid_leaf {
+            let fully_empty = if prev_valid_leaf.is_some() {
                 false
             } else {
                 let unbounded_ptr = self.unbounded_child.load(Relaxed, barrier);
@@ -707,12 +792,14 @@ where
                     if unbounded.retired() {
                         let deleted = unbounded.delete_self(Relaxed);
                         debug_assert!(deleted);
+
                         // It has to mark the pointer in order to prevent `LeafNode::insert` from
                         // replacing it with a new `Leaf`.
                         if let Some(obsolete_leaf) =
                             self.unbounded_child.swap((None, RETIRED), Release).0
                         {
                             barrier.reclaim(obsolete_leaf);
+                            uncleaned_leaf = true;
                         }
                         true
                     } else {
@@ -730,14 +817,15 @@ where
 
             drop(lock);
             if !self.has_retired_leaf(barrier) {
-                // Traverse several leaf nodes in order to cleanup deprecated links.
-                self.max_le_appr(removed_key, barrier);
-                return RemoveResult::Success;
+                break;
             }
         }
 
-        // Locking failed: give up expecting that the lock owner eventually cleans up the leaf.
-        RemoveResult::Success
+        if uncleaned_leaf {
+            RemoveResult::Cleanup
+        } else {
+            RemoveResult::Success
+        }
     }
 
     /// Checks if the [`LeafNode`] has a retired [`Leaf`].
@@ -817,6 +905,8 @@ where
     origin_leaf: AtomicArc<Leaf<K, V>>,
     low_key_leaf: AtomicArc<Leaf<K, V>>,
     high_key_leaf: AtomicArc<Leaf<K, V>>,
+    low_key_leaf_node: AtomicPtr<LeafNode<K, V>>,
+    high_key_leaf_node: AtomicPtr<LeafNode<K, V>>,
 }
 
 #[cfg(test)]
@@ -895,10 +985,9 @@ mod test {
                     leaf_node.rollback(&barrier);
                     for r in 0..(k - 1) {
                         assert_eq!(leaf_node.search(&r, &barrier), Some(&r));
-                        assert_eq!(
-                            leaf_node.remove_if::<_, _, false>(&r, &mut |_| true, &barrier),
-                            Ok(RemoveResult::Success)
-                        );
+                        assert!(leaf_node
+                            .remove_if::<_, _, false>(&r, &mut |_| true, &barrier)
+                            .is_ok());
                         assert_eq!(leaf_node.search(&r, &barrier), None);
                     }
                     assert_eq!(leaf_node.search(&(k - 1), &barrier), Some(&(k - 1)));
@@ -909,6 +998,9 @@ mod test {
                     assert_eq!(leaf_node.search(&(k - 1), &barrier), None);
                     break;
                 }
+                InsertResult::Retry(..) => {
+                    assert!(leaf_node.insert::<false>(k, k, &barrier).is_ok());
+                }
             }
         }
     }
@@ -948,7 +1040,7 @@ mod test {
                                         max_key.replace(id);
                                         break;
                                     }
-                                    InsertResult::Frozen(..) => {
+                                    InsertResult::Frozen(..) | InsertResult::Retry(..) => {
                                         continue;
                                     }
                                     _ => unreachable!(),
@@ -977,12 +1069,12 @@ mod test {
                                 &barrier,
                             ) {
                                 Ok(r) => match r {
-                                    RemoveResult::Success => break,
+                                    RemoveResult::Success | RemoveResult::Cleanup => break,
                                     RemoveResult::Fail => {
                                         assert!(removed);
                                         break;
                                     }
-                                    RemoveResult::Retired => unreachable!(),
+                                    RemoveResult::Frozen | RemoveResult::Retired => unreachable!(),
                                 },
                                 Err(r) => removed |= r,
                             }
@@ -1000,10 +1092,9 @@ mod test {
             for r in futures::future::join_all(task_handles).await {
                 assert!(r.is_ok());
             }
-            assert!(matches!(
-                leaf_node.remove_if::<_, _, false>(&usize::MAX, &mut |_| true, &Barrier::new()),
-                Ok(RemoveResult::Success | RemoveResult::Retired)
-            ));
+            assert!(leaf_node
+                .remove_if::<_, _, false>(&usize::MAX, &mut |_| true, &Barrier::new())
+                .is_ok());
         }
     }
 
